<?php

namespace FiloBlu\ExtInventory\Model\Algorithms;

use Magento\Framework\Exception\InputException;
use Magento\Framework\Exception\LocalizedException;
use Magento\InventoryApi\Api\GetSourcesAssignedToStockOrderedByPriorityInterface;
use Magento\InventorySourceSelectionApi\Api\Data\InventoryRequestInterface;
use Magento\InventorySourceSelectionApi\Api\Data\SourceSelectionItemInterface;
use Magento\InventorySourceSelectionApi\Api\Data\SourceSelectionItemInterfaceFactory;
use Magento\InventorySourceSelectionApi\Api\Data\SourceSelectionResultInterface;
use Magento\InventorySourceSelectionApi\Api\Data\SourceSelectionResultInterfaceFactory;
use Magento\InventorySourceSelectionApi\Model\GetInStockSourceItemsBySkusAndSortedSource;
use Magento\InventorySourceSelectionApi\Model\GetSourceItemQtyAvailableInterface;

/**
 * Class FinalSourceBasedAlgorithm
 * @package FiloBlu\ExtInventory\Model\Algorithms
 */
class DistanceFromShippingAddressAlgorithm extends AbstractAlgorithm
{
    /**
     * @var SourceSelectionResultInterfaceFactory
     */
    private $sourceSelectionResultFactory;
    /**
     * @var GetSourceItemQtyAvailableInterface
     */
    private $getSourceItemQtyAvailable;
    /**
     * @var GetInStockSourceItemsBySkusAndSortedSource
     */
    private $getInStockSourceItemsBySkusAndSortedSource;
    /**
     * @var SourceSelectionItemInterfaceFactory
     */
    private $sourceSelectionItemFactory;
    /**
     * @var GetSourcesAssignedToStockOrderedByPriorityInterface
     */
    private $getSourcesAssignedToStockOrderedByPriority;

    /**
     * FinalSourceBasedAlgorithm constructor.
     * @param SourceSelectionResultInterfaceFactory $sourceSelectionResultFactory
     * @param GetSourceItemQtyAvailableInterface $getSourceItemQtyAvailable
     * @param GetInStockSourceItemsBySkusAndSortedSource $getInStockSourceItemsBySkusAndSortedSource
     * @param SourceSelectionItemInterfaceFactory $sourceSelectionItemFactory
     * @param GetSourcesAssignedToStockOrderedByPriorityInterface $getSourcesAssignedToStockOrderedByPriority
     */
    public function __construct(
        SourceSelectionResultInterfaceFactory $sourceSelectionResultFactory,
        GetSourceItemQtyAvailableInterface $getSourceItemQtyAvailable,
        GetInStockSourceItemsBySkusAndSortedSource $getInStockSourceItemsBySkusAndSortedSource,
        SourceSelectionItemInterfaceFactory $sourceSelectionItemFactory,
        GetSourcesAssignedToStockOrderedByPriorityInterface $getSourcesAssignedToStockOrderedByPriority
    ) {
        $this->sourceSelectionResultFactory = $sourceSelectionResultFactory;
        $this->getSourceItemQtyAvailable = $getSourceItemQtyAvailable;
        $this->getInStockSourceItemsBySkusAndSortedSource = $getInStockSourceItemsBySkusAndSortedSource;
        $this->sourceSelectionItemFactory = $sourceSelectionItemFactory;
        $this->getSourcesAssignedToStockOrderedByPriority = $getSourcesAssignedToStockOrderedByPriority;
    }

    /**
     * @param InventoryRequestInterface $inventoryRequest
     * @return SourceSelectionResultInterface
     * @throws InputException
     * @throws LocalizedException
     */
    public function execute(InventoryRequestInterface $inventoryRequest): SourceSelectionResultInterface
    {
        $order = $this->getOrder($inventoryRequest);

        $countryId = $order->getShippingAddress()->getCountryId();
        $sourcesOptions = $this->getSources($inventoryRequest->getStockId(), $countryId, $inventoryRequest);

        foreach ($sourcesOptions as $option) {
            if ($option['fullOrder']) {
                return $this->sourceSelectionResultFactory->create(
                    [
                        'sourceItemSelections' => array_values($option['items']),
                        'isShippable'          => true
                    ]
                );
            }
        }

        $itemsLeft = [];
        foreach ($inventoryRequest->getItems() as $item) {
            $itemsLeft[$item->getSku()] = (float)$item->getQty();
        }

        // no full order create mixed order
        $selections = [];
        foreach ($sourcesOptions as $option) {
            /**
             * @var string $sku
             * @var SourceSelectionItemInterface $a
             */
            foreach ($option['items'] as $sku => $a) {
                if ($itemsLeft[$sku] <= 0.0) {
                    continue;
                }

                $qtyToDeduct = min($itemsLeft[$sku], $a->getQtyAvailable());

                $selections[] = $this->sourceSelectionItemFactory->create([
                    'sourceCode'   => $a->getSourceCode(),
                    'sku'          => $a->getSku(),
                    'qtyToDeduct'  => $qtyToDeduct,
                    'qtyAvailable' => $a->getQtyAvailable()
                ]);
                $itemsLeft[$sku] -= $qtyToDeduct;
            }
        }

        return $this->sourceSelectionResultFactory->create(
            [
                'sourceItemSelections' => $selections,
                'isShippable'          => true
            ]
        );
    }

    /**
     * @param $stockId
     * @param $countryId
     * @param $inventoryRequest
     * @return array
     * @throws InputException
     * @throws LocalizedException
     */
    protected function getSources($stockId, $countryId, $inventoryRequest)
    {
        $output = [];
        $sku = [];

        foreach ($inventoryRequest->getItems() as $item) {
            $sku[$item->getSku()] = (float)$item->getQty();
        }

        $sources = $this->getSourcesAssignedToStockOrderedByPriority->execute($stockId);

        foreach ($sources as $source) {
            $itemsInSource = $this->getInStockSourceItemsBySkusAndSortedSource->execute(array_keys($sku), [$source->getSourceCode()]);
            $result = [
                'source'     => $source->getSourceCode(),
                'country'    => $source->getCountryId(),
                'selections' => [],
                'fullOrder'  => count($sku) === count($itemsInSource)
            ];

            foreach ($itemsInSource as $stockItem) {
                $currentSku = $stockItem->getSku();

                if ($stockItem->getQuantity() < $sku[$currentSku]) {
                    $result['items'][$currentSku] = $this->sourceSelectionItemFactory->create([
                        'sourceCode'   => $source->getSourceCode(),
                        'sku'          => $currentSku,
                        'qtyToDeduct'  => $stockItem->getQuantity(),
                        'qtyAvailable' => $stockItem->getQuantity()
                    ]);
                    $result['fullOrder'] = false;
                } else {
                    $result['items'][$currentSku] = $this->sourceSelectionItemFactory->create([
                        'sourceCode'   => $source->getSourceCode(),
                        'sku'          => $currentSku,
                        'qtyToDeduct'  => $sku[$currentSku],
                        'qtyAvailable' => $sku[$currentSku]
                    ]);
                }
            }

            if ($source->getCountryId() === $countryId) {
                array_unshift($output, $result);
                continue;
            }
            $output[] = $result;
        }

        return $output;
    }
}
